------------------------------------------------------------------------
-- Boring lemmas used by the general lazy vectors
------------------------------------------------------------------------

module LazyVec.General.Lemmas where

open import Data.Function
open import Data.Nat
open import Data.Nat.Properties
open ℕ-semiringSolver
open import Data.Fin
open import Data.Vec
open import Data.Vec.Equality
open import Data.Vec.Properties
open import Relation.Binary.PropositionalEquality
open import Relation.Binary.HeterogeneousEquality

abstract

  lem₁ : forall m -> 3 + 3 * m ≡ 3 * suc m
  lem₁ m = prove (m ∷ [])
                 (con 3 :+ con 3 :* M)
                 (con 3 :* (con 1 :+ M))
                 ≡-refl
    where M = var zero

  lem₂ :  forall n₁ n₂
       -> n₁ + (1 + (n₁ + (1 + n₂))) ≡ n₁ + (1 + ((1 + n₁) + n₂))
  lem₂ n₁ n₂ = prove (n₁ ∷ n₂ ∷ [])
                     (N₁ :+ (con 1 :+ (N₁ :+ (con 1 :+ N₂))))
                     (N₁ :+ (con 1 :+ ((con 1 :+ N₁) :+ N₂)))
                     ≡-refl
    where N₁ = var zero; N₂ = var (suc zero)

  lem₃ :  forall {a n₁} (xs : Vec a n₁) n₂ y {n₃} (zs : Vec a n₃)
       -> xs ++ replicate {n = 1 + n₂} y ++ (y ∷ zs) ≅
          xs ++ replicate {n = 2 + n₂} y ++ zs
  lem₃ xs n₂ y zs = to-≅ $
    Vec-refl xs ++-cong replicate-lemma (1 + n₂) y zs

  lem₄ :  forall n₁ -> n₁ + (1 + n₁ + 0) ≡ n₁ + (1 + n₁)
  lem₄ n₁ = prove (n₁ ∷ [])
                  (N₁ :+ (con 1 :+ N₁ :+ con 0))
                  (N₁ :+ (con 1 :+ N₁))
                  ≡-refl
    where N₁ = var zero

  lem₅ :  forall {a n₁} (xs : Vec a n₁) {n₃} (ys : Vec a n₃)
       -> xs ++ (ys ++ []) ≅ xs ++ ys
  lem₅ xs ys = to-≅ $ Vec-refl xs ++-cong xs++[]=xs ys
