-- # import nat
-- # import vec
-- # import between
-- # import prelude
-- # import monoid
-- # import thunk
-- # import thunkedVec
-- # import thunkedMaybe

-- Digits with at least n₁ and at most n₂ elements, and actual length
-- len.

sig Digit (n₁, len, n₂ : ℕ) (a : Set) : Set where
  lenOK    : between n₁ len n₂
  elements : Vec len a

digit (n₁, len, n₂ : ℕ) (a : Set)
  :  (lenOK : between n₁ len n₂)
  -> (elements : Vec len a)
  -> Digit n₁ len n₂ a
digit lenOK elements = d
  where
  d : Digit _ _ _ _
  d = struct
    lenOK    = lenOK
    elements = elements

extendDigit (a : Set) (n₁, len, n₂ : ℕ)
  :  (n₁', n₂' : ℕ)
  -> Digit (n₁ + n₁') len  n₂        a
  -> Digit  n₁        len (n₂ + n₂') a
extendDigit n₁' n₂' d = struct
  lenOK    = limitBetweenLeft n₁' $ limitBetweenRight n₂' d.lenOK
  elements = d.elements

newLimits (a : Set) (n₁, n₁', len, n₂, n₂' : ℕ)
  :  between n₁' len n₂'
  -> Digit n₁  len n₂  a
  -> Digit n₁' len n₂' a
newLimits lenOK d = struct
  lenOK    = lenOK
  elements = d.elements

appendDigitSteps : ℕ -> ℕ
appendDigitSteps n₂ = n₂

appendDigit (a : Set) (n₁, len, n₂, n₁', len', n₂' : ℕ)
  :  Digit n₁ len n₂ a
  -> Digit n₁' len' n₂' a
  -> Thunk (appendDigitSteps n₂)
           (Digit (n₁ + n₁') (len + len') (n₂ + n₂') a)
appendDigit d₁ d₂ =
  return (digit $ betweenMon d₁.lenOK d₂.lenOK) `ap`
    slower (snd $ toLE d₁.lenOK)
           (d₁.elements ++′ d₂.elements)

consDigitSteps : ℕ
consDigitSteps = 0

consDigit (a : Set) (n₁, len, n₂ : ℕ)
  :  a
  -> Digit n₁ len n₂ a
  -> Thunk 0 (Digit (n₁ + 1) (len + 1) (n₂ + 1) a)
consDigit x d = return $ digit (bSS d.lenOK) $ x :: d.elements

snocDigitSteps : ℕ -> ℕ
snocDigitSteps n₂ = n₂

snocDigit (a : Set) (n₁, len, n₂ : ℕ)
  :  (d : Digit n₁ len n₂ a)
  -> a
  -> Thunk (snocDigitSteps n₂) (Digit (n₁ + 1) (len + 1) (n₂ + 1) a)
snocDigit d x = thunkM.do
  elems <- slower (snd $ toLE d.lenOK) (d.elements ++′ [x])
  return $ digit (bSS d.lenOK) elems

headDigitSteps : ℕ
headDigitSteps = 0

headDigit (a : Set) (n₁, len, n₂ : ℕ)
  : Digit (n₁ + 1) (len + 1) (n₂ + 1) a -> Thunk headDigitSteps a
headDigit d = tHead d.elements

lastDigitSteps : ℕ -> ℕ
lastDigitSteps n₂ = n₂

lastDigit (a : Set) (n₁, len, n₂ : ℕ)
  : Digit (n₁ + 1) (len + 1) (n₂ + 1) a -> Thunk (lastDigitSteps n₂) a
lastDigit d =
  slower (snd $ toLE $ betweenMonSubtract d.lenOK b₁₁₁) $
         tLast d.elements

tailDigitSteps : ℕ
tailDigitSteps = 0

tailDigit (a : Set) (n₁, len, n₂ : ℕ)
  :  Digit (n₁ + 1) (len + 1) (n₂ + 1) a
  -> Thunk tailDigitSteps (Digit n₁ len n₂ a)
tailDigit d =
  return (digit $ betweenMonSubtract d.lenOK b₁₁₁) `ap` tTail d.elements

initDigitSteps : ℕ -> ℕ
initDigitSteps n₂ = n₂

initDigit (a : Set) (n₁, len, n₂ : ℕ)
  :  Digit (n₁ + 1) (len + 1) (n₂ + 1) a
  -> Thunk (initDigitSteps n₂) (Digit n₁ len n₂ a)
initDigit d = slower (snd $ toLE lenOK) $
  return (digit lenOK) `ap` tInit d.elements
  where lenOK = betweenMonSubtract d.lenOK b₁₁₁

------------------------------------------------------------------------
-- Custom constructors

digit¹₁ (n : ℕ) (a : Set)
  : (x₁             : a) -> Digit 1 1 (n + 1) a
digit¹₁ x₁          = digit {1} betweenPlusPlus [x₁]

digit¹₂ (n : ℕ) (a : Set)
  : (x₁, x₂         : a) -> Digit 1 2 (n + 2) a
digit¹₂ x₁ x₂       = digit {1} betweenPlusPlus [x₁, x₂]

digit¹₃ (n : ℕ) (a : Set)
  : (x₁, x₂, x₃     : a) -> Digit 1 3 (n + 3) a
digit¹₃ x₁ x₂ x₃    = digit {1} betweenPlusPlus [x₁, x₂, x₃]

digit¹₄ (n : ℕ) (a : Set)
  : (x₁, x₂, x₃, x₄ : a) -> Digit 1 4 (n + 4) a
digit¹₄ x₁ x₂ x₃ x₄ = digit {1} betweenPlusPlus [x₁, x₂, x₃, x₄]
