-- Banker's queues based on rotate (from the real-time queues) instead
-- of reverse + append.

-- # import eq
-- # import nat
-- # import vec
-- # import thunk
-- # import order
-- # import cumSum
-- # import lazyVec
-- # import bankersQueueInvariants

data SRTBankersQueue (a : Set) : (Δ, L : ℕ) -> Set where
  BQ
    :  (Δ, L : ℕ)
    -> (Cs : Vec (L + Δ) ℕ)
    -> bqInvariant 0 (L + Δ) Δ Cs
    -> LazyVec a (L + Δ) Cs
    -> Vec L a
    -> SRTBankersQueue a Δ L

empty (a : Set) : SRTBankersQueue a 0 0
empty = BQ 0 0 [] (bqNil 0 0) lNil []

snocL' : (Δ, L : ℕ) -> ℕ
snocL' 0       L = 0
snocL' (suc _) L = L + 1

snocΔ' : (Δ, L : ℕ) -> ℕ
snocΔ' 0       L = 2 * L + 1
snocΔ' (suc Δ) L = Δ

-- I pretend that the arithmetic operations below take constant time,
-- and do not include them in the analysis. For the unary numbers
-- which I am using this is not true, but it should not be hard to
-- work around this with proper support for machine integers.

(▹) (Δ, L : ℕ) (a : Set)
  :  a
  -> SRTBankersQueue a Δ L
  -> Thunk 1 (SRTBankersQueue a (snocΔ' Δ L) (snocL' Δ L))
x ▹ BQ (suc Δ) L Cs inv f r = thunkM.do
  f' <- payLazyVec 1 f
  return' $ BQ Δ (L + 1) (Cs ⊖ 1) (bqSub {0} 1 inv) f' (x :: r)
x ▹ BQ 0       L Cs inv f r = thunkM.do
    f' <- rotate f (x :: r) lNil
    return $ BQ (2 * L + 1) 0
                (map suc Cs ++ replicate {L + 1} 0)
                (bqRaiseΔ {0} prop $ bqRot {0} inv)
                f' []
  where
  prop (n : ℕ) : n ≤ 2 * n + 1
  prop = mulIncreasingLeft {2} (leS leZ) `transitive` sucIncreasing

viewLL' : (Δ, L : ℕ) -> ℕ
viewLL' 0       L = 0
viewLL' (suc _) L = L

viewLΔ' (Δ, L : ℕ) : Δ ≥ 1 ∔ L ≥ 1 -> ℕ
viewLΔ' {0}     {0}     (left ge)  = absurdLE ge
viewLΔ' {0}     {0}     (right ge) = absurdLE ge
viewLΔ' {0}     {suc L} nonEmpty   = 2 * L + 1
viewLΔ' {suc Δ} {L}     nonEmpty   = Δ

viewL (Δ, L : ℕ) (a : Set)
  :  SRTBankersQueue a Δ L
  -> (nonEmpty : Δ ≥ 1 ∔ L ≥ 1)
  -> Thunk 2 (a × SRTBankersQueue a (viewLΔ' nonEmpty) (viewLL' Δ L))
viewL (BQ 0 0 _ _ _ _) (left ge)  = absurdLE ge
viewL (BQ 0 0 _ _ _ _) (right ge) = absurdLE ge
viewL (BQ (suc Δ) L (C :: Cs) inv f r) _ = thunkM.do
    f¹ <- payLazyVec 1 f
    f² <- slower prop $ lTail f¹
    return' $ pair (lHead f¹) $
      BQ Δ L (tail ((C :: Cs) ⊖ 1))
         (bqTail {0} inv)
         f² r
  where
  prop : C ∸ 1 ≤ 0
  prop = liftEq $ subtractLarger $ inv 1 (leS leZ)
viewL (BQ 0 (suc L) (C :: Cs) inv f r) _ = thunkM.do
    f¹ <- payLazyVec 1 f
    f² <- slower prop $ lTail f¹
    f³ <- rotate f² r lNil
    return $ pair (lHead f¹) $
      BQ (2 * L + 1) 0
         (map suc (Cs ⊖ (1 ∸ C)) ++ replicate {L + 1} 0)
         (bqRaiseΔ {0} (plusIncreasingRight {L + 1}) $
            bqRot {0} $ bqTail {0} {_} {0} inv)
         f³ []
  where prop : C ∸ 1 ≤ 0
        prop = smallerRelated `transitive` inv 1 (leS leZ)
