-- Monolithic vector operations

-- # import nat
-- # import vec
-- # import thunk
-- # import thunkedPair
-- # import thunkedUnit

vecRec (a : Set) (m, n : ℕ)
  :   (P : (n : ℕ) -> Vec n a -> Set)
  ->  Thunk m (P 0 [])
  ->  ((n : ℕ) -> (x : a) -> (xs : Vec n a) -> P n xs -> Thunk m (P (n + 1) (x :: xs)))
  ->  (xs : Vec n a)
  ->  Thunk ((n + 1) * (m + 1)) (P n xs)
vecRec _ n c []        = slower' {1} n
vecRec _ n c (x :: xs) = step $
  substThunk (cong ((+) 1) addCommutative) refl' $
    c _ x xs =<< vecRec _ n c xs

tReverse (a : Set) (m : ℕ)
  : Vec m a -> Thunk m (Vec m a)
tReverse xs = tReverse' xs []
  where
  tReverse' (a : Set) (m, n : ℕ)
    : Vec m a -> Vec n a -> Thunk m (Vec (m + n) a)
  tReverse' nil        ys = return ys
  tReverse' (x :: xs') ys = step $ tReverse' xs' (x :: ys)

infixr 5 ++′

(++′) (a : Set) (m, n : ℕ)
  : Vec m a -> Vec n a -> Thunk m (Vec (m + n) a)
[]        ++′ ys = return ys
(x :: xs) ++′ ys = step $ return ((::) x) `ap` (xs ++′ ys)

tHead (a : Set) (n : ℕ)
  : Vec (n + 1) a -> Thunk 0 a
tHead (x :: _) = return x

tTail (a : Set) (n : ℕ)
  : Vec (n + 1) a -> Thunk 0 (Vec n a)
tTail (_ :: xs) = return xs

tInit (a : Set) (n : ℕ)
  : Vec (n + 1) a -> Thunk n (Vec n a)
tInit [_]            = return []
tInit (x :: y :: ys) = step $ return ((::) x) `ap` tInit (y :: ys)

tLast (a : Set) (n : ℕ)
  : Vec (n + 1) a -> Thunk n a
tLast [x]              = return x
tLast (_ :: (x :: xs)) = step $ tLast (x :: xs)

tFoldr (steps, n : ℕ) (a, b : Set)
  :  (a -> b -> Thunk steps b)
  -> b
  -> Vec n a
  -> Thunk (n * (steps + 1)) b
tFoldr                 f e []        = return e
tFoldr {steps} {suc n} f e (x :: xs) = step $
  substThunk addCommutative refl' $
    f x =<< tFoldr f e xs

tEnumFrom : (ℓ, from : ℕ) -> Thunk ℓ (Vec ℓ ℕ)
tEnumFrom 0       _    = return []
tEnumFrom (suc ℓ) from = step $
  return ((::) from) `ap` tEnumFrom ℓ (suc from)
