------------------------------------------------------------------------
-- Erasing thunks
------------------------------------------------------------------------

-- Functions converting thunked things to the corresponding simple
-- things, with all thunks and library primitives erased.

module Erasure where

open import Kind
open import TypeSystem

open import Relation.Binary.PropositionalEquality
  using (_≡_; ≡-refl; ≡-subst; ≡-cong)

-- Erases types.

⌜_⌝⋆ : Ty thunked -> Ty simple
⌜ Nat ⌝⋆       = Nat
⌜ σ × τ ⌝⋆     = ⌜ σ ⌝⋆ × ⌜ τ ⌝⋆
⌜ σ → τ ⌝⋆     = ⌜ σ ⌝⋆ → ⌜ τ ⌝⋆
⌜ Thunk n τ ⌝⋆ = ⌜ τ ⌝⋆

-- Thunks inside a pay context are erased, so we get this lemma:

erase-lemma :  forall {m n τ} C
            -> ⌜ C [ Thunk m τ ] ⌝⋆ ≡ ⌜ C [ Thunk n τ ] ⌝⋆
erase-lemma ●            = ≡-refl
erase-lemma (const● σ)   = ≡-refl
erase-lemma (C ●× σ)     = ≡-cong (\τ -> τ × ⌜ σ ⌝⋆) (erase-lemma C)
erase-lemma (σ ×● C)     = ≡-cong (\τ -> ⌜ σ ⌝⋆ × τ) (erase-lemma C)
erase-lemma (Thunk● n C) = erase-lemma C

-- Erases contexts.

⌜_⌝▻ : Ctxt thunked -> Ctxt simple
⌜ ε ⌝▻     = ε
⌜ Γ ▻ σ ⌝▻ = ⌜ Γ ⌝▻ ▻ ⌜ σ ⌝⋆

-- Type signature for erasers of term-like things.

TmEraser : TmLike -> Set
TmEraser _•_ = forall {Γ τ} -> Γ • τ -> ⌜ Γ ⌝▻ • ⌜ τ ⌝⋆

-- Erases variables.

⌜_⌝∋ : TmEraser _∋_
⌜ vz ⌝∋                   = vz
⌜ vs x ⌝∋                 = vs ⌜ x ⌝∋
⌜ gwait∋ {Γ = Γ} C m x ⌝∋ =
  ≡-subst (\τ -> ⌜ Γ ⌝▻ ∋ τ) (erase-lemma C) ⌜ x ⌝∋

-- Erases terms.

⌜_⌝ : TmEraser _⊢_
⌜ var x ⌝               = var ⌜ x ⌝∋
⌜ λ t ⌝                 = λ ⌜ t ⌝
⌜ λ² t ⌝                = λ² ⌜ t ⌝
⌜ t₁ · t₂ ⌝             = ⌜ t₁ ⌝ · ⌜ t₂ ⌝
⌜ fix t ⌝               = fix ⌜ t ⌝
⌜ z ⌝                   = z
⌜ s ⌝                   = s
⌜ natrec t₁ t₂ ⌝        = natrec ⌜ t₁ ⌝ ⌜ t₂ ⌝
⌜ (t₁ , t₂) ⌝           = (⌜ t₁ ⌝ , ⌜ t₂ ⌝)
⌜ uncurry t ⌝           = uncurry ⌜ t ⌝
⌜ return t ⌝            = ⌜ t ⌝
⌜ t₁ >>= t₂ ⌝           = ⌜ t₂ ⌝ · ⌜ t₁ ⌝
⌜ √ t ⌝                 = ⌜ t ⌝
⌜ pay n₂ t ⌝            = ⌜ t ⌝
⌜ gpay {Γ = Γ} C n₂ t ⌝ =
  ≡-subst (\τ -> ⌜ Γ ⌝▻ ⊢ τ) (erase-lemma C) ⌜ t ⌝
⌜ gwait {Γ = Γ} C n₂ t ⌝ =
  ≡-subst (\τ -> ⌜ Γ ⌝▻ ⊢ τ) (erase-lemma C) ⌜ t ⌝
