------------------------------------------------------------------------
-- Lazy evaluation of simple terms
------------------------------------------------------------------------

-- This module is not part of the formalisation, it is only included
-- to give a different view of the operational semantics. (I have not
-- formally proved that the evaluation function defined here is
-- equivalent to the operational semantics, though.)

module Evaluation.Simple where

open import Data.Nat
open import Kind
open import TypeSystem
open import Substitution
open VarSubst
open import Value
import Heap
open Heap _⊢_

infix 15 ⟦_⟧_
infix 10 _⊢⇓ⁿ_

-- The result of evaluating a term.

data _⊢⇓ⁿ_ (Γ : Ctxt simple) (τ : Ty simple) : Set where
  ⇓ : forall {Δ} -> Γ ⇒ Δ -> Heap Δ -> Δ ⊢⇓ τ
      -- Number of steps necessary to evaluate it.
      -> ℕ
      -> Γ ⊢⇓ⁿ τ

-- Evaluation. Note that this function is partial, due to the presence
-- of general fixpoints.

⟦_⟧_ : forall {Γ τ} -> Γ ⊢ τ -> Heap Γ -> Γ ⊢⇓ⁿ τ

-- Variables.

⟦ var vz     ⟧ Σ₁ ▷ʰ t with ⟦ t ⟧ Σ₁
... | ⇓ ρ Σ₂ v n = ⇓ (ρ ↑) (Σ₂ ▷ʰ valToTm v) (v /⊢⇓ wk) n
⟦ var (vs v) ⟧ Σ₁ ▷ʰ t with ⟦ var v ⟧ Σ₁
... | ⇓ ρ Σ₂ val n = ⇓ (ρ ↑) (Σ₂ ▷ʰ t /⊢ ρ) (val /⊢⇓ wk) n
⟦ var ()     ⟧ ∅ʰ

-- Values.

⟦ λ t ⟧                         Σ₁ = ⇓ idˢ        Σ₁                     (λ⇓ t)                      0
⟦ λ² t ⟧                        Σ₁ = ⇓ idˢ        Σ₁                     (λ²⇓ t)                     0
⟦ z ⟧                           Σ₁ = ⇓ idˢ        Σ₁                     z⇓                          0
⟦ s ⟧                           Σ₁ = ⇓ idˢ        Σ₁                     s⇓                          0
⟦ natrec {σ = σ} t₁ t₂ ⟧        Σ₁ = ⇓ wk         (Σ₁ ▷ʰ t₁)             (natrec⇓ vz (t₂ /⊢ wk ↑ ↑)) 0
⟦ _,_ {σ = σ₁} {τ = σ₂} t₁ t₂ ⟧ Σ₁ = ⇓ (wk ∘ˢ wk) (Σ₁ ▷ʰ t₁ ▷ʰ t₂ /⊢ wk) (vs vz ,⇓ vz)               0
⟦ uncurry t ⟧                   Σ₁ = ⇓ idˢ        Σ₁                     (uncurry⇓ t)                0

-- Fix.

⟦ fix t ⟧ Σ₁ with ⟦ t ⟧ (Σ₁ ▷ʰ fix t)
... | ⇓ ρ Σ₂ v n = ⇓ (wk ∘ˢ ρ) Σ₂ v (1 + n)

-- Applications.

⟦ t₁ · t₂ ⟧ Σ₁ with ⟦ t₁ ⟧ Σ₁

  -- Successor.

⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ Σ₂ s⇓ n = ⇓ (wk⇒ ρ) (Σ₂ ▷ʰ t₂ /⊢ ρ) (s₁⇓ vz) n

  -- Lambda.

⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ₁ Σ₂ (λ⇓ t₁') n₁ with ⟦ t₁' ⟧ (Σ₂ ▷ʰ t₂ /⊢ ρ₁)
... | ⇓ ρ₂ Σ₃ v n₂ = ⇓ (wk⇒ ρ₁ ∘ˢ ρ₂) Σ₃ v (suc (n₁ + n₂))
⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ Σ₂ (λ²⇓ t₁') n = ⇓ (wk⇒ ρ) (Σ₂ ▷ʰ t₂ /⊢ ρ) (λ⇓ t₁') n

  -- Natrec.

⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ₁ Σ₂ (natrec⇓ v₁¹ t₁²) n₁ with ⟦ t₂ /⊢ ρ₁ ⟧ Σ₂

    -- Zero.

⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ₁ Σ₂ (natrec⇓ v₁¹ t₁²) n₁
               | ⇓ ρ₂ Σ₃ z⇓ n₂ with ⟦ var v₁¹ /⊢ ρ₂ ⟧ Σ₃
... | ⇓ ρ₃ Σ₄ v n₃ = ⇓ (ρ₁ ∘ˢ ρ₂ ∘ˢ ρ₃) Σ₄ v (suc (n₁ + n₂ + n₃))

    -- Suc.

⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ₁ Σ₂ (natrec⇓ v₁¹ t₁²) n₁
               | ⇓ ρ₂ Σ₃ (s₁⇓ v₂') n₂
               with ⟦ t₁² /⊢ ρ₂ ↑ ↑ /⊢ sub v₂' ↑ ⟧
                    (Σ₃ ▷ʰ (natrec (var v₁¹) t₁² /⊢ ρ₂) · var v₂')
... | ⇓ ρ₃ Σ₄ v n₃ = ⇓ (wk⇒ (ρ₁ ∘ˢ ρ₂) ∘ˢ ρ₃) Σ₄ v (suc (n₁ + n₂ + n₃))

  -- Uncurry.

⟦ t₁ · t₂ ⟧ Σ₁ | ⇓ ρ₁ Σ₂ (uncurry⇓ t₁') n₁ with ⟦ t₂ /⊢ ρ₁ ⟧ Σ₂
... | ⇓ ρ₂ Σ₃ (v₂¹ ,⇓ v₂²) n₂ with ⟦ t₁' /⊢ ρ₂ ↑ ↑ /⊢ sub v₂¹ ↑ /⊢ sub v₂² ⟧ Σ₃
...   | ⇓ ρ₃ Σ₄ v n₃ = ⇓ (ρ₁ ∘ˢ ρ₂ ∘ˢ ρ₃) Σ₄ v (suc (n₁ + n₂ + n₃))
